import numpy as np
from matplotlib import pyplot as plt
from matplotlib import patches
        
from icenumerics.geometry import *
from icenumerics import ureg

class spin():
    """ 
    A spin is defined by two vectors in R3 space.
    The vector center gives the position of the center of the spin
    The vector direction gives the dipole moment of the spin.
    """
    
    def __init__(self,center,direction):
        
        self.center = np.array(center)*center.units
        self.direction = np.array(direction)*center.units
        
    def __str__(self):
        return("Spin with Center at [%d %d %d] and Direction [%d %d %d]\n" %\
               (tuple(self.center)+tuple(self.direction)))

    def display(self,ax1):
        
        X=self.center[0].magnitude
        Y=self.center[1].magnitude
        DX=self.direction[0].magnitude*0.3
        DY=self.direction[1].magnitude*0.3
        W = np.sqrt(DX**2+DY**2)
        ax1.plot([X],[Y],'b')
        ax1.add_patch(patches.Arrow(X-DX,Y-DY,2*DX,2*DY,width=W,fc='b'))

class spins(list): 
    """ `spins` is a very general class that contains a list of spin objects. The only feature of this list is that it is created from the centers and directions of the spins, and also that it contains a ´display´ method. """ 
    
    def __init__(self, centers = [], directions = None):
        """To initialize, we can give the centers and directions of the spins contained. However, we can also initialize an empty list, and then populate it using the `extend` method """
        
        if len(centers)>0:
            self = self.extend([spin(c,d) for (c,d) in zip(centers,directions)])
        
        
    def display(self,ax = None, ix = False):
        """ This displays the spins in a pyplot axis. The ix parameter allows us to obtain the spins index, which is useful to access specific indices."""
        if not ax:
            fig1, ax = plt.subplots(1,1)            

        for s in self:
            s.display(ax)

        plt.axis("equal")

    def create_lattice(self, geometry, size,
        lattice_constant = 1, border = "closed spin"):
        """ Creates a lattice of spins. 
        The geometry can be:
            * "square"
            * "honeycomb"
        The border can be 
            * "closed spin":
            * "closed vertex" 
            * "periodic" 
        """
        
        latticeunits = lattice_constant.units

        if geometry == "square":
            center, direction = square_spin_ice_geometry(
                size[0], size[1], lattice_constant.magnitude,
                "Random", Ratio = 1, Boundary = border
            )
                    
        elif geometry == "honeycomb":
            center, direction = honeycomb_spin_ice_geometry(
                size[0], size[1], lattice_constant.magnitude,
                "Random"
            )
            
        else: 
            raise(ValueError(geometry+" is not a supporteed geometry."))
        
        self.__init__(center*latticeunits,direction*latticeunits)
        self.lattice = lattice_constant
        self.ordering = "random"




import numpy as np
import copy as cp
import matplotlib.pyplot as plt
import matplotlib.patches as ptch
import scipy.spatial as spa
import pandas as pd

from icenumerics.geometry import *
from icenumerics.parameters import *
from icenumerics.spins import *

from icenumerics.magcolloids  import magcolloid as mc

class colloid_in_trap():
    """
    An object ColloidInTrap represents a colloidal particle in a bistable trap. 
    It has three main properties:
    * center is the center of the bistable trap. 
    * direction is a vector (whose magnitude is not important) that points from one stable
        position to the other.
    * colloid is a vector that indicates where the colloid is placed with respect to the center.
    
    Each of these quantities are represented by lists of three elements, which correspond to 
    vectors in R3 space. 
    
    A colloid_in_trap object also has the properties:
    * colloid_properties:
        * susceptibility
        * diffusion
        * diameter
        * rel_density
        * volume
    * trap_properties
        * trap_sep: The distance between the traps.
        * height
        * stiffness
        * spread
    """
    def __init__(self, 
        center, direction, particle, trap):
        """ initializes a colloid_in_trap object """
        
        self.center = np.array(center)*center.units
        
        # Direction is always unitary
        self.direction = np.array(direction)/np.linalg.norm(direction.magnitude,2)
        
        self.particle = particle
        self.trap = trap
                
        """ Colloid position is given by the direction vectors"""
        self.colloid = self.direction * self.trap.trap_sep/2
        
    def __str__(self):
        """ Prints a string which represents the colloid_in_trap """
        return("Colloid is in [%d %d %d], trap is [%d %d %d %d %d %d]\n" %\
               (tuple(self.colloid.magnitude)+tuple(self.center.magnitude)+tuple(self.direction)))
               
    def display(self, ax1=False):
        """ Draws a figure with the trap and the colloid inside it"""
        if not ax1:
            fig, ax1 = plt.subplots(1,1)
            
        X=self.center[0].magnitude
        Y=self.center[1].magnitude
        # D is the vector that goes fom the center to each of the traps
        DX=self.direction[0]/2*self.trap.trap_sep.to(self.center[0].units).magnitude
        DY=self.direction[1]/2*self.trap.trap_sep.to(self.center[1].units).magnitude
        # P is the vector that goes from the center to the colloid
        PX=self.colloid[0].to(self.center[0].units).magnitude
        PY=self.colloid[1].to(self.center[1].units).magnitude
        
        #Discriminant = self.colloid.dot(self.direction)
        #Discriminant = Discriminant/abs(Discriminant)
        
        #DX = DX*Discriminant
        #DY = DY*Discriminant
                
        W = (1/self.trap.stiffness.magnitude)*4e-3
        
        ax1.plot(X,Y,'k')
        ax1.add_patch(patches.Circle(
            (X-DX,Y-DY),radius = W,
            ec='g', fc='g'))
        ax1.add_patch(patches.Circle(
            (X+DX,Y+DY),radius = W,
            ec='y', fc='y'))
        ax1.add_patch(patches.Circle(
            (X+PX,Y+PY), radius = W/3, ec='k', fc = 'none'))
        #ax1.plot([X,X+PX],[Y,Y+PY],color='k')
        
    def flip(self):
        """flips the ColloidInTrap by inverting its direction and its colloid attributes. Returns fliped object"""
        cp = copy.deepcopy(self);
        
        cp.direction = self.direction*(-1)
        cp.colloid = self.colloid*(-1)
        return cp
        
    def bias(self, vector):
        """ 
        Flips the ColloidInTrap to make it point in the direction of vector (dot(colloid,vector)>0). Returns fliped object        
        """

        # if vector is 2D, append zero
        if len(vector)==2:
            if vector.__class__.__name__=="tuple":
                vector = vector+0
            if vector.__class__.__name__=="ndarray":
                vector = np.hstack((vector,0))
            if vector.__class__.__name__=="list":
                vector = vector+[0]
        elif len(vector)>3:
            raise ValueError("The vector argument has to be 2 or 3 dimentions")
        
        cp = copy.deepcopy(self);

        if np.array(vector).dot(self.direction)<0:
            cp = self.flip()
        return cp
        
class colloidal_ice(list):
    """ 
    The colloidal ice object is a list of colloid_in_trap objects.
    It also includes some extra parameters contained in the worldparams attribute. 
    It normally takes a spin ice object as input and generates one colloid_in_trap object for each spin
    """
    def __init__(self,
    arrangement,particle,trap,
    height_spread = 0, susceptibility_spread = 0,
    region = None, periodic = None):
        """ 
        The arrangement parameter defines the positions and directions of the colloidal ice. There are two possible inputs:
            * a `spins` object: in this case the colloidal ice is copied from the spins arrangement. 
            * a `dict` object: this `dict` object must contain two arrays, `center` and `direction`.
        `particle` and `trap` are parameter containers created with the `particle` and `trap` generators. They can be a single object, or a list. If it is a list, it must coincide with the number of elements defined by the `arrangement` parameter.
        """
        
        if arrangement.__class__.__name__ == "spins":
            centers = [s.center for s in arrangement]
            directions = [s.direction for s in arrangement]
            
        else:
            centers = arrangement['centers']
            directions = arrangement['directions']
        
        if not hasattr(particle,'__getitem__'):
            particle = [particle for c in centers]
        if not hasattr(trap,'__getitem__'):
            trap = [trap for c in centers]
        
        self.height_spread=height_spread
        self.susceptibility_spread = susceptibility_spread
                    
        self.extend(
            [colloid_in_trap(c,d,p,t) 
                for p,t,c,d in zip(particle,trap,centers,directions)])

        if region == None:
            units = centers[0].units
            lower_bounds = np.min(
                np.array([c.to(units).magnitude for c in centers])*units,0)
            upper_bounds = np.max(
                np.array([c.to(units).magnitude for c in centers])*units,0)
            
            region = np.vstack([lower_bounds,upper_bounds])*units
        
        self.region = region
        
        if periodic is None:
            periodic = False
            
        self.periodic = periodic
        
                
    def display(self, ax = None):
                
        if not ax:
            fig1, ax = plt.subplots(1,1)            

        for s in self:
            s.display(ax)
        
        ax.set_xlim([self.region[0,0].magnitude,self.region[1,0].magnitude])
        ax.set_ylim([self.region[0,1].magnitude,self.region[1,1].magnitude])
         
        ax.set_aspect("equal")
           
        #plt.axis("square")
    
    def pad_region(self,pad):
        self.region[0] = self.region[0]-pad
        self.region[1] = self.region[1]+pad
        
    def simulate(self, world, name,
        targetdir = '',
        include_timestamp = True,
        run_time = 60*ureg.s,
        framerate = 15*ureg.Hz,
        timestep = 100*ureg.us,
        output = ["x","y","z"]):
        
        particles = [c.particle for c in self]
        traps = [c.trap for c in self]

        colloids = [c.colloid.to(ureg.um).magnitude for c in self]*ureg.um
        centers = [c.center.to(ureg.um).magnitude for c in self]*ureg.um
        directions = [c.direction for c in self]
        
        # s = np.shape(np.array(colloids))
        # initial_randomization = np.random.randn(s[0],s[1])*0.1*ureg.um
        initial_displacement = np.array([[0,0,0.1]]*len(colloids))*ureg.um
        
        p_type = np.array(classify_objects(particles))
        t_type = np.array(classify_objects(traps))

        for p in np.unique(p_type):

            particles = mc.particles(
                colloids+centers+initial_displacement,
                atom_type = 0,
                atoms_id = np.arange(len(colloids)),
                radius = self[p].particle.radius,
                susceptibility = self[p].particle.susceptibility,
                drag = self[p].particle.drag)

        for t in np.unique(t_type):

            traps = mc.bistable_trap(
                centers,
                directions,
                particles,
                atom_type = 1, 
                trap_id = np.arange(len(centers))+len(colloids),
                distance = self[t].trap.trap_sep,
                height = self[t].trap.height,
                stiffness = self[t].trap.stiffness,
                height_spread = self.height_spread
                )
                
        world_sim = mc.world(
            particles,traps,
            region=self.region.transpose().flatten(),
            walls=[False,False,False],
            temperature = world.temperature,
            dipole_cutoff = world.dipole_cutoff,
            lj_cutoff = 0,
            lj_parameters = [0*ureg.pg*ureg.um**2/ureg.us**2,0],
            enforce2d = world.enforce2d,
            gravity = 0*ureg.um/ureg.us**2)
            
        field = mc.field(magnitude = world.field, 
            frequency = 0*ureg.Hz, angle = 0*ureg.degrees)
        
        self.sim = mc.sim(
            file_name = name, dir_name = targetdir, stamp_time = include_timestamp,
            timestep = timestep,
            framerate = framerate,
            total_time = run_time,
            output = output,
            particles = particles, traps = traps, world = world_sim, field = field)
        
        self.sim.generate_scripts()
        self.sim.run()
        
        self.trj = self.sim.load(read_trj = True)
        self.frames = self.trj.index.get_level_values("frame").unique()
        self.set_state_from_frame(frame = -1)
        
    def set_state_from_frame(self, frame):
        
        idx = pd.IndexSlice
        frame = self.frames[frame]
        for i,c in enumerate(self):
            c.colloid = self.trj.loc[idx[frame,i+1],["x","y","z"]].values*ureg.um - c.center
            c.direction = c.direction * np.sign(np.dot(c.colloid,c.direction))
    
    def calculate_energy(self):
        """ Calculates the sum of the inverse cube of all the inter particle distances.
        For this it uses the spatial package of scipy which shifts this calculation to
        a compiled program and it's therefore faster.
        The energy output is given in 1/nm^3
        """
        colloids = np.array([np.array(c.center+c.colloid) for c in self])
        self.energy = sum(spa.distance.pdist(colloids)**(-3))
        return self.energy
        
def classify_objects(object_list):
    """ Classifies objects by uniqueness. Returns a list with an object type directory."""
    o_type = -1

    def where(obj,obj_list):
        """returns the first occurence of `particle` in the array `particles`"""
        for i,o in enumerate(obj_list):
            if o==obj:
                return i

    type_dict = []

    for i,o in enumerate(object_list):
        loc = where(o,object_list[0:i])
        if loc is not None:
            type_dict.append(o_type)
        else:
            o_type = o_type+1
            type_dict.append(o_type)
    
    return type_dict


from icenumerics.spins import *
from icenumerics.colloidalice import colloidal_ice
import subprocess # Subprocess is a default library which allows us to call a command line program from within a python script
import shutil # shutil allows us to move files around. This is usefull to organize the resulting input and output files. 
import os
import sys

import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import scipy.spatial as sptl
import pandas as pd

def spin_crossing_point(S1,S2):
    # This works well in 2d. In 3d it's triciker
    if not (S1['Direction']==S2['Direction']).all():
        A = np.ones([2,2])
        A[:,0] = S1['Direction']
        A[:,1] = -S2['Direction']

        b = np.array([
            S2['Center'][0]-S1['Center'][0],
            S2['Center'][1]-S1['Center'][1]])

        lam = np.linalg.solve(A,b)

        return S1['Center']+lam[0]*S1['Direction']
    else:
        return np.Inf+np.zeros(np.shape(S1['Center']))

def unique_points(Points,Tol = 0.1):
    """Returns only the distinct points (with a tolerance)."""
    Distance = sptl.distance.squareform(sptl.distance.pdist(Points))
    IsLast = []
    for i,p in enumerate(Points):
        IsLast = IsLast + [not np.any(Distance[(i+1):,i]<=Tol)]
        
    return Points[np.array(IsLast),:]
    
def colloidal_ice_vector(C):
    """Extracts an array of centers and directions from a Colloidal Ice System"""
    Vectors = np.array(np.zeros(len(C)),dtype=[('Center',np.float,(2,)),('Direction',np.float,(2,))])
    i=0
    for c in C:
        Vectors[i] = (c.center[0:2],c.direction[0:2])
        i=i+1
    return Vectors
        
def calculate_neighbor_pairs(Centers):
    """This function makes a list of all the Pairs of Delaunay Neighbors from an array of points"""
    
    tri = sptl.Delaunay(Centers)

    # List all Delaunay neighbors in the system
    NeighborPairs = np.array(np.zeros(2*np.shape(tri.simplices)[0]),
                             dtype=[('Pair',np.int,(2,)),('Distance',np.float),('Vertex',np.float,(2,))])

    i = 0
    for t in tri.simplices:
        NeighborPairs[i]['Pair'] = np.sort(t[0:2])
        NeighborPairs[i]['Distance'] = sptl.distance.euclidean(Centers[t[0]],Centers[t[1]])
        NeighborPairs[i+1]['Pair'] = np.sort(t[1:3])
        NeighborPairs[i+1]['Distance'] = sptl.distance.euclidean(Centers[t[1]],Centers[t[2]])
        i = i+2

    return NeighborPairs

def from_neighbors_get_nearest_neighbors(NeighborPairs):
    # This function takes a list of Delaunay Neighbor Pairs and returns only those which are close to the minimum distance.
    NeighborPairs['Distance']=np.around(NeighborPairs['Distance'],decimals=4)
    NeighborPairs = NeighborPairs[NeighborPairs['Distance']<=np.min(NeighborPairs['Distance'])*1.1]
    
    return NeighborPairs

def get_vertices_positions(NeighborPairs,spins):
    # From a list of Spins, get neighboring spins, and get the crossing point of each, which defines a vertex.  
    for i,n in enumerate(NeighborPairs):
        NeighborPairs[i]['Vertex'] = spin_crossing_point(spins[n['Pair'][0]],spins[n['Pair'][1]])[0:2]
    
    return NeighborPairs

class vertices():
    def __init__(self):
        # This function initializes the Vertices Array
        self.array=np.array([],
                      dtype=[
                          ('Location', float,(2,)),
                          ('id',int),
                          ('Coordination',int),
                          ('Charge',int),
                          ('Dipole',int,(2,))])
        
    def colloids_to_vertices(self,C):

        self.spins = colloidal_ice_vector(C)
        
        NeighborPairs = calculate_neighbor_pairs(self.spins['Center'])

        NeighborPairs = from_neighbors_get_nearest_neighbors(NeighborPairs)

        NeighborPairs = get_vertices_positions(NeighborPairs,self.spins)
        
        v = unique_points(NeighborPairs['Vertex'])
        
        ## Make Vertex array
        self.array=np.array(np.empty(np.shape(v)[0]),
                            dtype=[
                                ('Location', float,(2,)),
                                ('id',int),
                                ('Coordination',int),
                                ('Charge',int),
                                ('Dipole',int,(2,))])
        
        self.array['Location'] = v
        
        ## Make Neighbors directory
        self.neighbors = {}

        for i,v in enumerate(self.array):
            v['id']=i
            self.neighbors[i] = []
            for n in NeighborPairs:
                if sptl.distance.euclidean(n['Vertex'],v['Location'])<np.mean(n['Vertex']*1e-6):
                    self.neighbors[i]=self.neighbors[i]+list(n['Pair'])
            self.neighbors[i] = set(self.neighbors[i])
            
            ## Calculate Coordination
            v['Coordination'] = len(self.neighbors[i])
        
            ## Calculate Charge and Dipole
            v['Charge'] = 0
            v['Dipole'] = [0,0]
               
            for n in self.neighbors[v['id']]:
                v['Charge']=v['Charge'] + np.sign(np.sum((v['Location']-self.spins[n]['Center'])*self.spins[n]['Direction']))

                v['Dipole']=v['Dipole'] + self.spins[n]['Direction']
         
        return self
        
    def display(self,ax = False,DspCoord = False):
                
        if not ax:
            fig1, ax = plt.subplots(1,1)  

        if not DspCoord:
            for v in self.array:
                if v['Charge']>0:
                    c = 'r'
                else:
                    c = 'b'
                ax.add_patch(patches.Circle(
                    (v['Location'][0],v['Location'][1]),radius = abs(v['Charge'])*2,
                    ec='none', fc=c))
                X = v['Location'][0]
                Y = v['Location'][1]
                if v['Charge']==0:
                    DX = v['Dipole'][0]
                    DY = v['Dipole'][1]
                    ax.add_patch(patches.Arrow(X-DX,Y-DY,2*DX,2*DY,width=5,fc='k'))
                
        if DspCoord: 
            for v in self.array:
                if v['Charge']>0:
                    c = 'r'
                else:
                    c = 'b'
                    
                ax.add_patch(patches.Circle(
                    (v['Location'][0],v['Location'][1]),radius = abs(v['Coordination'])*2,
                    ec='none', fc=c))
                X = v['Location'][0]
                Y = v['Location'][1]
        
        #ax.set_aspect("equal")    
        #plt.axis("equal")




import numpy as np
import warnings
from . import ureg

class trap():
    # The trap geometry object contains:
    # TrapSep: Distance between stable points of the traps [in nm]
    # Height: Hill height of the trap [in nm]
    # Stiffness: Stiffness of the trap [in pNnm].
    
    # Trap Separation could be determined also by the spin dipole.
    # I need to think how to put them together. 

    def __init__(self,
        trap_sep = 10e3*ureg.nm,
        height = 200*ureg.nm,
        stiffness = 1.2e-4 * ureg.pN/ureg.nm):
        """ This initializes a trap parameter set. """
        
        self.trap_sep = trap_sep;
        self.height = height;
        self.stiffness = stiffness;

class particle():
      def __init__(self, 
                radius = 2*ureg.um,
                susceptibility = 1,
                drag = 4e6*ureg.pN/(ureg.um/ureg.s), 
                diffusion = None, temperature = None, 
                density = 1000*ureg.kg/ureg.m**3):
        """
        Initializes a particle type. 
        The diffusion coefficient can be given instead o the drag.
        In that case, the temperature is also needed to calculatelm the drag. 
        This represents the density mismatch of the particle in the solvent
        """
        
        if diffusion:
            
            kB = (1.38064852e-23*ureg.J/ureg.K).to(ureg.pN*ureg.nm/ureg.K)
            
            KbT = kB*temperature
            drag = KbT/diffusion
            
        self.radius = radius.to(ureg.um)
        self.susceptibility = susceptibility
        self.drag = drag.to(ureg.pg/ureg.us)
        self.mass = (density*4/3*np.pi*(radius)**3).to(ureg.pg)
        
        damp = 1e-3*ureg.us
        
        self.drag_mass = (drag*damp).to(ureg.pg)
        
class world():
    
    def __init__(self,
            temperature = 300*ureg.K,
            field = 20*ureg.mT,
            dipole_cutoff = 200*ureg.um,
            enforce2d = True):
        """ The force parameter adds a biasing force to the simulation."""
        
        self.temperature = temperature
        self.kB = (1.38064852e-23*ureg.J/ureg.K).to(ureg.pN*ureg.nm/ureg.K)
        self.enforce2d = enforce2d
        self.field = field
        self.dipole_cutoff = dipole_cutoff



from pint import UnitRegistry

import sys

from .magcolloids import magcolloid as mc

ureg = mc.ureg

from icenumerics.spins import *
from icenumerics.colloidalice import *
from icenumerics.vertices import *


import numpy as np

def SquareSpinIceConcatenateCenters(VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice,Boundary):
    """ Concatenate Vertical and Horizontal spin arrays """
    if Boundary == "closed spin":
        Delta = 0.5
    elif Boundary == "closed vertex":
        Delta = -0.5

    return np.vstack(
    (
        np.hstack(
            (
                VSpinsX, # X
                VSpinsY+Delta*Lattice, #Y 
                np.zeros(VSpinsX.shape,float) #Z
                ) # Centers of Vertical Spins
            ), 
         np.hstack(
            (
                HSpinsX+Delta*Lattice, #X
                HSpinsY, #Y
                np.zeros(HSpinsX.shape,float) #Z
                ) # Centers of Horizontal Spins
            )
        ) # (Centers Vertical, Centers Horizontal)
    ) # Centers of All Spins, Vertical stacked on top of Horizontal
    return Center

def SquareSpinIceDirection(VOrderDirectionArray,HOrderDirectionArray):
    return np.vstack(
    (
        np.hstack(
            (
                np.zeros(VOrderDirectionArray.shape,float),
                VOrderDirectionArray,
                np.zeros(VOrderDirectionArray.shape,float)
                )
            ),
         np.hstack(
            (
                HOrderDirectionArray,
                np.zeros(HOrderDirectionArray.shape,float),
                np.zeros(HOrderDirectionArray.shape,float)
                )
            )
        ) # (Directions Vertical, Directions Horizontal)
    ) # Direction of All Spins

def SquareSpinIceDirectionGSOrdering(VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice):
    VOrderDirectionArray = np.ones(VSpinsX.shape,float)
    VOrderDirectionArray[
        np.mod(
            VSpinsX+VSpinsY,
            2*Lattice)>0]=-1

    HOrderDirectionArray = np.ones(HSpinsX.shape,float)
    HOrderDirectionArray[
        np.mod(
            HSpinsX+HSpinsY,
            2*Lattice)==0]=-1
    return VOrderDirectionArray, HOrderDirectionArray

def SquareSpinIceDirectionRandomOrdering(VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice):
    VOrderDirectionArray = np.random.rand(*VSpinsX.shape)
    VOrderDirectionArray[VOrderDirectionArray>0.5]=1;
    VOrderDirectionArray[VOrderDirectionArray<=0.5]=-1;
    
    HOrderDirectionArray = np.random.rand(*HSpinsX.shape)
    HOrderDirectionArray[HOrderDirectionArray>0.5]=1;
    HOrderDirectionArray[HOrderDirectionArray<=0.5]=-1;

    return VOrderDirectionArray, HOrderDirectionArray

def square_spin_ice_geometry(Sx,Sy,Lattice,Ordering,Ratio,Boundary):
    """ 
    This function calculates the positions and directions of the spins in a square spin ice system. 
    
    It does so by creating two meshes, 
    where one mesh defines the position of the horizontal spins, 
    and the other defines the position of the vertical spins.
    Then a subroutine concatenates the arrays.
    """
    
    vsoX = np.arange(0,Sx+1)*Lattice
    vsoY = np.arange(0,Sy+1)*Lattice
    
    if Boundary == "closed spin":
        VSpinsX, VSpinsY = np.meshgrid(vsoX,vsoY[0:-1])
        HSpinsX, HSpinsY = np.meshgrid(vsoX[0:-1],vsoY)
    elif Boundary == "closed vertex":
        VSpinsX, VSpinsY = np.meshgrid(vsoX[0:-1],vsoY)
        HSpinsX, HSpinsY = np.meshgrid(vsoX,vsoY[0:-1])
    else:
        raise(ValueError(Boundary+" is not a valid Boundary specification"))


    VSpinsX = VSpinsX.flatten(1)[None].T
    VSpinsY = VSpinsY.flatten(1)[None].T
    HSpinsX = HSpinsX.flatten(1)[None].T
    HSpinsY = HSpinsY.flatten(1)[None].T

    Center = SquareSpinIceConcatenateCenters(VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice,Boundary)

    if Ordering == "Random":
        VOrderDirectionArray, HOrderDirectionArray = \
            SquareSpinIceDirectionRandomOrdering(
                VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice)

    elif Ordering == "GroundState":
        
        VOrderDirectionArray, HOrderDirectionArray = \
            SquareSpinIceDirectionGSOrdering(
                VSpinsX,VSpinsY,HSpinsX,HSpinsY,Lattice)
            
    elif Ordering == "Biased":
        VOrderDirectionArray = np.ones(VSpinsX.shape,float)
        HOrderDirectionArray = np.ones(HSpinsX.shape,float)

    else:
        error("I do not know this ordering")
    
    Direction = SquareSpinIceDirection(VOrderDirectionArray*Ratio,HOrderDirectionArray)
    
    Direction = Direction*Lattice
    return Center, Direction



import numpy as np
import random
import scipy.spatial as spa

def HoneycombSpinIceDirectionRandomOrdering(Direction):

    Direction = np.array([dir*random.randrange(-1,2,2) for dir in Direction])
    return Direction

def HoneycombSpinIceDirectionSpinSolidPhase(Center,Lattice,Direction):
    """ Honeycomb Spin Solid Phase """
    sqrt3 = np.sqrt(3)
    epsilon = 10**-6

    #To improve readibility, the quantities on this should be given names that are descriptive of their function. 
    
    # JJ1 are the vertical spins in even rows (starting from zero). One over 3 should all point down: those in 3*Lattice/2, 9*Lattice/2,  12*Lattice/2.
   
    JJ1= np.logical_or(
        (Center[:,1]/Lattice) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon)

    Direction[JJ1]=abs(Direction[JJ1])

    JJ12 = np.logical_and(JJ1,np.logical_or
                           ((Center[:,0]/Lattice -3/2) % 3/2 < epsilon, 
                            abs((Center[:,0]/Lattice -3/2) % 3/2 -3/2)<epsilon)) 

    Direction[JJ12]=-Direction[JJ12]

    # JJ2 are the tilted spins below the even rows. The 3 first (starting from 0) should point right, the 3 following should point left, ect..

    JJ2 = np.logical_or(
        ((Center[:,1]/Lattice)+ (sqrt3/4)) % (sqrt3) < epsilon, 
        abs((Center[:,1]/Lattice +sqrt3/4) % sqrt3 - sqrt3) < epsilon)

    # JJ5 are the spins in JJ2 that should point right.

    JJ5=np.logical_and(JJ2,(((((4*Center[:,0])/Lattice)-1 + epsilon)//2)//3 % 2) < epsilon)
        

    JJ7=np.logical_and(JJ5,Direction[:,0]>0)

    Direction[JJ7]=-Direction[JJ7]

    # JJ9 are the spins in JJ2 that should point left.


    JJ9=np.logical_and(JJ2,np.logical_not((((((4*Center[:,0])/Lattice)-1 + epsilon)//2)//3 % 2) < epsilon))

    JJ11=np.logical_and(JJ9,Direction[:,0]<0)

    Direction[JJ11]=-Direction[JJ11]

    # JJ4 are vertical spins in odd rows. One over 3 should all point down: those in 0, 3*Lattice,  6*Lattice.

    JJ4 = np.logical_or(
        ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon)

    Direction[JJ4]=abs(Direction[JJ4])

    JJ42 = np.logical_and(JJ4, (Center[:,0]/Lattice) % 3 < epsilon)

    Direction[JJ42]=-Direction[JJ42]

    # JJ3 are the tilted spins above the even rows. 
    JJ3 = np.logical_or(
        ((Center[:,1]/Lattice) - (sqrt3/4)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice - sqrt3/4) % sqrt3 - sqrt3) < epsilon)

    JJ6=np.logical_and(JJ3,(((((4*Center[:,0])/Lattice)-1 + epsilon)//2)//3 % 2) < epsilon)

    JJ8=np.logical_and(JJ3,Direction[:,0]<0)

    Direction[JJ8]=-Direction[JJ8]

    JJ10=np.logical_and(JJ3,np.logical_not(((((((4*Center[:,0])/Lattice)-1 + epsilon)//2)//3 % 2) < epsilon)))
     
    JJ12=np.logical_and(JJ10,Direction[:,0]>0)

    Direction[JJ12]=-Direction[JJ12]

    return Direction

def HoneycombSpinIceDirectionBandAntiferromagnetic(Center,Lattice,Direction):
    """This state has lines in one direction in a ferromagnetic state with alternating orientation, and bands in the other directions in an antiferromagnetic state."""
    sqrt3 = np.sqrt(3)
    epsilon = 10**-6
    
    # JJ1 are the vertical spins in even rows (starting from zero). These should all point up. 
    JJ1= np.logical_or(
        (Center[:,1]/Lattice) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon)


    Direction[JJ1,1]=abs(Direction[JJ1,1])

   # JJ2 are the tilted spins below the even rows. These should all point right
    JJ2 = np.logical_or(
        ((Center[:,1]/Lattice)+ (sqrt3/4)) % (sqrt3) < epsilon, 
        abs((Center[:,1]/Lattice +sqrt3/4) % sqrt3 - sqrt3) < epsilon)
    
    # JJ5 are the spins in JJ2 that point left. These we flip.
    JJ5 = np.logical_and(JJ2, Direction[:,0]<0)
    Direction[JJ5] = -Direction[JJ5]
    
    # JJ3 are the tilted spins above the even rows. These should all point left
    JJ3 = np.logical_or(
        ((Center[:,1]/Lattice) - (sqrt3/4)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice - sqrt3/4) % sqrt3 - sqrt3) < epsilon)
    
    # JJ7 are the spins in JJ3 that point right. These we flip
    JJ7= np.logical_and(JJ3,Direction[:,0]>0)
    Direction[JJ7]=-Direction[JJ7]

    # JJ4 are vertical spins in odd rows. These should all point down.
    JJ4 = np.logical_or(
        ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon)

    Direction[JJ4]=-abs(Direction[JJ4])    
    
    return Direction

def HoneycombSpinIceDirectionBandMixedAntiferromagnetic(Center,Lattice,Direction):
    """The state has the bands in one direction in a ferromagnetic state, and the bands in the other two directions in an antiferromagnetic state."""
    
    sqrt3 = np.sqrt(3)
    epsilon = 10**-6

    JJ1= np.logical_and( 
         np.logical_or(
        (Center[:,1]/Lattice) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon),
         np.logical_or(
        (Center[:,0]/Lattice -1/2) % (2) < epsilon,
        abs((Center[:,0]/Lattice -1/2))% (2) < epsilon))


    Direction[JJ1,1]=abs(Direction[JJ1,1])

    JJ11= np.logical_and( 
          np.logical_or(
         (Center[:,1]/Lattice) % (sqrt3) < epsilon,
         abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon),
          np.logical_or(
         (Center[:,0]/Lattice +1/2) % (2) < epsilon,
         abs((Center[:,0]/Lattice +1/2))% (2) < epsilon))

    Direction[JJ11,1]=-abs(Direction[JJ11,1])

     # JJ2 are the tilted spins below the even rows. These should all point right
    JJ2 = np.logical_or(
        ((Center[:,1]/Lattice)+ (sqrt3/4)) % (sqrt3) < epsilon, 
        abs((Center[:,1]/Lattice +sqrt3/4) % sqrt3 - sqrt3) < epsilon)
    
    # JJ5 are the spins in JJ2 that point left. These we flip.
    JJ5 = np.logical_and(JJ2, Direction[:,0]<0)
    Direction[JJ5] = -Direction[JJ5]
    
    # JJ3 are the tilted spins above the even rows. These should all point left
    JJ3 = np.logical_or(
        ((Center[:,1]/Lattice) - (sqrt3/4)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice - sqrt3/4) % sqrt3 - sqrt3) < epsilon)
    
    # JJ7 are the spins in JJ3 that point right. These we flip
    JJ7= np.logical_and(JJ3,Direction[:,0]>0)
    Direction[JJ7]=-Direction[JJ7]

   # JJ4 are vertical spins in odd rows. These should all point down.
    JJ4 = np.logical_and(
        np.logical_or(
            ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
            abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon),
        np.logical_or(
            (Center[:,0]/Lattice) % (2) < epsilon,
            abs((Center[:,0]/Lattice))% (2) < epsilon))

    Direction[JJ4]=-abs(Direction[JJ4])    

    JJ44 = np.logical_and( 
           np.logical_or(
          ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
           abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon),
           np.logical_or(
          (Center[:,0]/Lattice +1) % (2) < epsilon,
           abs((Center[:,0]/Lattice +1)% (2) < epsilon)))

    Direction[JJ44]=abs(Direction[JJ44])

    return Direction    

def HoneycombSpinIceDirectionFerromagnetic(Center,Lattice,Direction):
    """In the ferromagnetic state all particles are oriented in one direction"""
    sqrt3 = np.sqrt(3)
    epsilon = 10**-6
        
    JJ1= np.logical_and( 
     np.logical_or(
        (Center[:,1]/Lattice) % (sqrt3) < epsilon,
         abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon),
     np.logical_or(
        (Center[:,0]/Lattice -1/2) % (2) < epsilon,
         abs((Center[:,0]/Lattice -1/2))% (2) -(2) < epsilon))
                                            

    Direction[JJ1,1]=abs(Direction[JJ1,1])

    JJ11= np.logical_and( 
     np.logical_or(
         (Center[:,1]/Lattice) % (sqrt3) < epsilon,
          abs((Center[:,1]/Lattice)% (sqrt3) - sqrt3) < epsilon),
     np.logical_or(
         (Center[:,0]/Lattice +1/2) % 2 < epsilon,
          abs((Center[:,0]/Lattice +1/2)% 2 - 2) < epsilon))

    Direction[JJ11,1]=abs(Direction[JJ11,1])

     # JJ21 are the tilted spins below the even rows, on the left side of the cell. These should all point right.
    JJ21 =np.logical_and( 
      np.logical_or(
         ((Center[:,1]/Lattice)+ (sqrt3/4)) % (sqrt3) < epsilon, 
          abs((Center[:,1]/Lattice +sqrt3/4) % sqrt3 - sqrt3) < epsilon),
      np.logical_or(
         (Center[:,0]/Lattice -1/4) % (2) < epsilon,
          abs((Center[:,0]/Lattice -1/4))% (2) < epsilon))

   
    # JJ5 are the spins in JJ21 that point left. These we flip.
    JJ5 = np.logical_and(JJ21, Direction[:,1]<0)
    Direction[JJ5] = -Direction[JJ5]

    # JJ22 are the tilted spins below the even rows, on the right side of the cell. These should all point left.
    JJ22 =np.logical_and( 
      np.logical_or(
         ((Center[:,1]/Lattice)+ (sqrt3/4)) % (sqrt3) < epsilon, 
          abs((Center[:,1]/Lattice +sqrt3/4) % sqrt3 - sqrt3) < epsilon),
      np.logical_or(
         (Center[:,0]/Lattice -3/4) % (2) < epsilon,
          abs((Center[:,0]/Lattice -3/4))% (2) < epsilon))
    
    # JJ52 are the spins in JJ22 that point right. These we flip.
    JJ52 = np.logical_and(JJ22, Direction[:,1]<0)
    Direction[JJ52] = -Direction[JJ52]
    
    # JJ31 are the tilted spins above the even rows, on the left side of the cell. These should all point right.
    JJ31 =np.logical_and(
      np.logical_or(
        ((Center[:,1]/Lattice) - (sqrt3/4)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice - sqrt3/4) % sqrt3 - sqrt3) < epsilon),
      np.logical_or(
         (Center[:,0]/Lattice -1/4) % (2) < epsilon,
          abs((Center[:,0]/Lattice -1/4))% (2) < epsilon))
    
    # JJ7 are the spins in JJ3 that point left. These we flip
    JJ7= np.logical_and(JJ31,Direction[:,1]<0)
    Direction[JJ7]=-Direction[JJ7]

   # JJ32 are the tilted spins above the even rows, on the left side of the cell. These should all point right.
    JJ32 =np.logical_and(
      np.logical_or(
        ((Center[:,1]/Lattice) - (sqrt3/4)) % (sqrt3) < epsilon,
        abs((Center[:,1]/Lattice - sqrt3/4) % sqrt3 - sqrt3) < epsilon),
      np.logical_or(
         (Center[:,0]/Lattice -3/4) % (2) < epsilon,
          abs((Center[:,0]/Lattice -3/4))% (2) < epsilon))
    
    # JJ72 are the spins in JJ32 that point right. These we flip
    JJ72= np.logical_and(JJ31,Direction[:,1]<0)
    Direction[JJ72]=-Direction[JJ72]
    

   # JJ4 are vertical spins in odd rows. These should all point down.
    JJ4 = np.logical_and( 
        np.logical_or(
           ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
           abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon),
        np.logical_or(
          (Center[:,0]/Lattice) % (2) < epsilon,
           abs((Center[:,0] / Lattice)% 2 - 2) < epsilon))

    Direction[JJ4,1]=abs(Direction[JJ4,1])    

    JJ44 = np.logical_and( 
           np.logical_or(
          ((Center[:,1]/Lattice)+ (sqrt3/2)) % (sqrt3) < epsilon,
           abs((Center[:,1]/Lattice + sqrt3/2) % sqrt3 - sqrt3) < epsilon),
           np.logical_or(
          (Center[:,0]/Lattice +1) % 2 < epsilon,
           abs((Center[:,0]/Lattice +1)% (2) - 2) < epsilon))

    Direction[JJ44,1]=abs(Direction[JJ44,1])

    return Direction
    
def honeycomb_spin_ice_geometry(Sx,Sy,Lattice,Ordering):
    """This function calculates the positions and directions of the spins in a honeycomb spin ice system. 

    These are the arrays to iterate. For now, each point in x-y generates one unit cell which is a hexagon of spins. Then repeated spins are eliminated."""
    """ 
    Include here an explanation of each of the ordering options"""
    
    x = np.arange(0,Sx) * Lattice
    y = np.arange(0,Sy) * Lattice
    t = np.arange(0,2*np.pi,np.pi/3)

    CenterX = np.mod(x+y[:,np.newaxis]*np.cos(np.pi/3),Sx*Lattice) + \
        Lattice*np.cos(t.reshape(t.size,1,1))*0.5
    CenterY = np.zeros(x.shape)+y[:,np.newaxis]*np.sin(np.pi/3) + \
        Lattice*np.sin(t.reshape(t.size,1,1))*0.5

    Center = np.vstack([CenterX.flatten(), \
        CenterY.flatten(), \
        np.zeros(CenterX.flatten().shape)]).T

    DirectionX = np.zeros(x.shape) + np.zeros(y[:,np.newaxis].shape) + \
        (np.cos(t.reshape(t.size,1,1)))
    
    DirectionY = np.zeros(x.shape) + np.zeros(y[:,np.newaxis].shape) - \
        (np.sin(t.reshape(t.size,1,1)))

    Direction = np.vstack([DirectionY.flatten(), \
        DirectionX.flatten(), \
        np.zeros(DirectionX.flatten().shape)]).T

    """This erases repeated spins"""
    """
    For this we find all neighbors within a small tolerance (using cKDTree is fast).
    We then make an array of all ids to remove by listing only the second member of each neighbor pair.
    We make a mask with the remove array and apply it to the arrays Center and Direction """
    tree = spa.cKDTree(Center)
    remove = [p[1] for p in tree.query_pairs(1e-10)]
    
    mask = np.ones(len(Center),dtype=bool)
    mask[remove] = False
    
    Center = Center[mask]
    Direction = Direction[mask]
    
    if Ordering == "Random":
        Direction = HoneycombSpinIceDirectionRandomOrdering(Direction)

    elif Ordering == "SpinSolid":
        Direction = HoneycombSpinIceDirectionRandomOrdering(Direction)
    elif Ordering == "Ferromagnetic":
        Direction = HoneycombSpinIceDirectionFerromagnetic(Center,Lattice,Direction)
    elif Ordering == "BandMixedAntiferromagnetic":
        Direction = HoneycombSpinIceDirectionBandMixedAntiferromagnetic(Center,Lattice,Direction)
    elif Ordering == "BandAntiferromagnetic":
        Direction = HoneycombSpinIceDirectionBandAntiferromagnetic(Center,Lattice,Direction)
    else:
        error("I do not know this ordering")
    
#    for c in Center:
#        c = RoundToN(c,6)
    Direction = Direction*Lattice
    return Center, Direction




import setuptools

with open("Readme.md","r") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name="icenumerics",
    version="0.1.1",
    author="Antonio Ortiz-Ambriz",
    author_email="aortiza@gmail.com",
    description="Simulations of Colloidal Ice",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aortiza/icenumerics",
    packages=setuptools.find_packages(),
    install_requires = ['numpy','scipy','pandas','matplotlib','ureg','jsonpickle','magcolloid'],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
    ),
)